from typing import List
from pvpower.estimator import Estimator
from pvpower.traindata import TrainData



def test(estimator: Estimator, train_data: TrainData):
    return TestReport(estimator, train_data)



class TestReport:

    def __init__(self, estimator: Estimator, train_data: TrainData):
        example_data, validation_data = train_data.split()
        self.validation_samples  = [reocrd for reocrd in validation_data.samples]
        estimator.retrain(example_data)
        self.predictions = [estimator.predict(validation_record) for validation_record in self.validation_samples]

    def __percent(self, real, predicted):
        if real == 0:
            return 0
        elif predicted == 0:
            return 0
        else:
            return round((predicted * 100 / real) - 100, 2)

    def __diff_all(self) -> List[float]:
        diff_total = []
        for i in range(len(self.validation_samples)):
            if self.validation_samples[i].irradiance == 0:
                continue
            else:
                predicted = self.predictions[i]

            real = self.validation_samples[i].power_watt
            if real == 0:
                if predicted == 0:     # ignore true 0 predictions to void wasting the score
                    diff = 0
                else:
                    diff = 100
            else:
                if real < 10 and abs(predicted - real) < 10:  # ignore diff < 10
                    diff = 0
                else:
                    diff = self.__percent(real, predicted)
            diff_total.append(diff)
        return sorted(diff_total)

    @property
    def score(self) -> float:
        values = sorted(list(self.__diff_all()))
        values = values[2:-2]
        abs_values = [abs(value) for value in values]
        if len(abs_values) == 0:
            return 10000
        else:
            return round(sum(abs_values) / len(abs_values), 2)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        txt = "score:  " + str(self.score) + " (samples: " + str(len(self.validation_samples)) + ")\n"
        txt += '{:14s}        {:14s} {:14s} {:14s} {:14s} {:14s}         {:10s} {:10s}           {:10s}\n'.format("time", "irradiance", "sunshine", "cloud_cover", "visibility", "proba.fog", "real", "predicted", "diff[%]")
        for i in range(0, len(self.validation_samples)):
            txt += '{:<14s}        {:<14d} {:<14d} {:<14d} {:<14d}  {:<14d}        {:<10d} {:<10d}           {:<10d}\n'.format(self.validation_samples[i].time.strftime("%d.%b  %H:%M"),
                                                                                                                               self.validation_samples[i].irradiance,
                                                                                                                               self.validation_samples[i].sunshine,
                                                                                                                               self.validation_samples[i].cloud_cover,
                                                                                                                               self.validation_samples[i].visibility,
                                                                                                                               self.validation_samples[i].probability_for_fog,
                                                                                                                               self.validation_samples[i].power_watt,
                                                                                                                               self.predictions[i],
                                                                                                                               int(self.__percent(self.validation_samples[i].power_watt, self.predictions[i])))
        txt = txt + "\nscore:  " + str(self.score)
        return txt



import logging
import pytz
from datetime import datetime
from pvpower.mosmix import MemoryCachedMosmixLoader
from typing import Optional


class WeatherForecast:

    def __init__(self,
                 time: datetime,
                 irradiance: int,
                 sunshine: int,
                 cloud_cover: int,
                 probability_for_fog: int,
                 visibility: int):
        self.time_utc = time.astimezone(pytz.UTC)
        self.irradiance = irradiance
        self.sunshine = sunshine
        self.cloud_cover = cloud_cover
        self.probability_for_fog = probability_for_fog
        self.visibility = visibility

    @property
    def time(self) -> datetime:
        return datetime.strptime((self.time_utc + (datetime.now() - datetime.utcnow())).strftime("%d.%m.%Y %H:%M:%S"), "%d.%m.%Y %H:%M:%S")

    def with_time(self, dt: datetime):
        return WeatherForecast(dt,
                               self.irradiance,
                               self.sunshine,
                               self.cloud_cover,
                               self.probability_for_fog,
                               self.visibility)

    def __str__(self):
        return self.time_utc.isoformat() + " (" + self.time.isoformat() + ")" + \
               ", irradiance=" + str(round(self.irradiance)) + \
               ", sunshine=" + str(round(self.sunshine)) + \
               ", cloud_cover=" + str(round(self.cloud_cover)) + \
               ", probability_for_fog=" + str(round(self.probability_for_fog)) + \
               ", visibility=" + str(round(self.visibility))


class WeatherStation:

    def __init__(self, station: str):
        self.__mosmix_loader = MemoryCachedMosmixLoader(station)

    def forcast_from(self) -> datetime:
        return self.__mosmix_loader.get().date_from

    def forcast_to(self) -> datetime:
        return self.__mosmix_loader.get().date_to

    def forecast(self, time: datetime = None) -> Optional[WeatherForecast]:
        time = time if time is not None else datetime.now()

        mosmix = self.__mosmix_loader.get()
        if mosmix.supports(time):
            forecast = WeatherForecast(time,
                                       round(mosmix.rad1h(time)),
                                       round(mosmix.sund1(time)),
                                       round(mosmix.neff(time)),
                                       round(mosmix.wwm(time)),
                                       round(mosmix.vv(time)))
            return forecast
        else:
            logging.info("forecast record for " + time.strftime("%Y.%m.%d %H:%M") + " not available. Returning None (current mosmix: " + str(mosmix) + ")")
            return None



from datetime import datetime, timedelta
from typing import List, Dict, Optional
from pvpower.forecast import PvPowerForecast, LabelledWeatherForecast



class TimeFrame:

    def __init__(self, hourly_forecasts: List[LabelledWeatherForecast]):
        self.__hourly_forecasts = hourly_forecasts

    @property
    def start_time(self) -> datetime:
        return self.__hourly_forecasts[0].time

    @property
    def end_time(self) -> datetime:
        return self.__hourly_forecasts[-1].time + timedelta(minutes=59)

    @property
    def width_hours(self) -> int:
        return len(self.__hourly_forecasts)

    @property
    def hourly_power(self) -> List[int]:
        return [hourly_forecast.power_watt for hourly_forecast in self.__hourly_forecasts]

    @property
    def power_total(self) -> int:
        return sum(self.hourly_power)

    def __lt__(self, other):
        return self.start_time < other.start_time

    def __eq__(self, other):
        return self.start_time == other.start_time

    def __repr__(self):
        return self.__str__()

    def __str__(self) -> str:
        return self.start_time.strftime("%H:%M") + "->" + self.end_time.strftime("%H:%M") + "; expected power/hour: " + ", ".join([str(hourly_forecast.power_watt)  + " watt" for hourly_forecast in self.__hourly_forecasts])



class TimeFrames:

    @staticmethod
    def of(frames: List[TimeFrame]):
        return  TimeFrames(sorted(frames, key=lambda frame: "{0:10d}".format(1000000-frame.power_total) + frame.start_time.strftime("%d.%m.%Y %H"), reverse=False))

    def __init__(self, frames: List[TimeFrame]):
        self.__frames = frames

    def filter(self, min_watt_per_hour: int):
        filtered_frames = [frame for frame in self.__frames if max(sorted(frame.hourly_power)) > min_watt_per_hour]
        full_frames = [frame for frame in filtered_frames if (len([power for power in frame.hourly_power if power >= min_watt_per_hour]) == len(frame.hourly_power))]
        if len(full_frames) > 0:
            non_overlapping_full_frames = [full_frames[0]]
            for frame in full_frames:
                if frame.start_time > non_overlapping_full_frames[-1].end_time:
                    non_overlapping_full_frames.append(frame)

            frames = [frame for frame in filtered_frames if frame not in non_overlapping_full_frames]
            frames = non_overlapping_full_frames + frames
            return TimeFrames(frames)
        else:
            return TimeFrames(filtered_frames)

    def empty(self) -> bool:
        return len(self.__frames) == 0

    def __len__(self):
        return len(self.__frames)

    def best(self) -> Optional[TimeFrame]:
        if len(self.__frames) > 0:
            return self.__frames[0]
        else:
            return None

    def pos(self, pos: int, dflt: TimeFrame = None) -> TimeFrame:
        if dflt is None:
            dflt = self.__frames[0]
        if pos < len(self.__frames):
            return self.__frames[pos]
        else:
            return dflt

    def all(self) -> List[TimeFrame]:
        return list(self.__frames)

    def __str__(self):
        txt = "start time ......... end time ........... pv power total ........... power per hour\n"
        for frame in sorted(self.__frames, key=lambda frame: frame.start_time, reverse=False):
            power = str(round(frame.power_total))
            power_per_hour = ", ".join([str(round(power)) for power in frame.hourly_power])
            txt += frame.start_time.strftime("%d %b, %H:%M") + "  ..... " + \
                   frame.end_time.strftime("%d %b, %H:%M") + " " + \
                   "".join(["."] * (15 - len(power))) + " " + power + " watt" + " " + \
                   "".join(["."] * (25 - len(power_per_hour))) + " " + power_per_hour + "\n"
        return txt



class Next24hours:

    def __init__(self, pv_forecast: PvPowerForecast, predicted_power: Dict[datetime, LabelledWeatherForecast]):
        self.__predicted_power = predicted_power
        self.__pv_forecast = pv_forecast

    @staticmethod
    def __round_hour(dt: datetime) -> datetime:
        return (dt.replace(second=0, microsecond=0, minute=0, hour=dt.hour) +timedelta(hours=dt.minute//30))

    @staticmethod
    def of(pv_forecast: PvPowerForecast):
        now = datetime.strptime((datetime.now()).strftime("%d.%m.%Y %H") + ":00", "%d.%m.%Y %H:%S")
        predicted_power = {}
        for weather_forecast in [pv_forecast.weather_forecast_service.forecast(prediction_time) for prediction_time in [now + timedelta(hours=i) for i in range(0, 40)]]:
            if weather_forecast is not None:
                predicted_value = pv_forecast.predict_by_weather_forecast(weather_forecast)
                if predicted_value is not None:
                    predicted_power[Next24hours.__round_hour(weather_forecast.time)] = LabelledWeatherForecast.create(weather_forecast, predicted_value)
        return Next24hours(pv_forecast, predicted_power)

    def __prediction_values(self) -> List[int]:
        return [forecast.power_watt for forecast in self.__predicted_power.values() if forecast.time <= (datetime.now() + timedelta(hours=24))]

    def peek(self) -> int:
        return max(self.__prediction_values())

    def peek_time(self) -> datetime:
        peek_time = None
        peek_value = 0
        for dt in self.__predicted_power.keys():
            forecast = self.__predicted_power[dt]
            if forecast.power_watt > peek_value:
                peek_value = forecast.power_watt
                peek_time = dt
        return peek_time

    def power_total(self) -> int:
        return sum(self.__prediction_values())

    def frames(self, width_hours: int = 1) -> TimeFrames:
        frames = []
        times = list(self.__predicted_power.keys())
        for offset_hour in range(0, 24+width_hours):
            forecasts = [self.__predicted_power[times[idx]] for idx in range(offset_hour, offset_hour + width_hours)]
            frame = TimeFrame(forecasts)
            frames.append(frame)
        frames = [frame for frame in frames if frame.start_time <= (datetime.now() + timedelta(hours=24))]
        return TimeFrames.of(frames)

    def __str__(self):
        txt = "time ................ pv power ..... irradiance ....... sunshine .... visibility .... fog probab. .... cloud cover\n"
        for time in list(self.__predicted_power.keys())[:24]:
            power = str(round(self.__predicted_power[time].power_watt))
            irradiance = str(round(self.__predicted_power[time].irradiance))
            visibility = str(round(self.__predicted_power[time].visibility))
            sunshine = str(round(self.__predicted_power[time].sunshine))
            probability_for_fog = str(round(self.__predicted_power[time].probability_for_fog))
            cloud_cover = str(round(self.__predicted_power[time].cloud_cover))
            txt += time.strftime("%d %b, %H:%M") + " " + \
                   "".join(["."] * (10 - len(power))) + " " + power + " watt " + \
                   "".join(["."] * (15 - len(irradiance))) + " " + irradiance + " " + \
                   "".join(["."] * (15 - len(sunshine))) + " " + sunshine + \
                   "".join(["."] * (15 - len(visibility))) + " " + visibility + " " + \
                   "".join(["."] * (15 - len(probability_for_fog))) + " " + probability_for_fog +  " " + \
                   "".join(["."] * (15 - len(cloud_cover))) + " " + cloud_cover + "\n"
        txt += str(self.__pv_forecast) + "\n"
        return txt





import unittest
from datetime import datetime, timedelta
from pvpower.weather_forecast import WeatherStation


class TestWeatherForecast(unittest.TestCase):

    def test_date_is_not_timezone_aware(self):
        station = WeatherStation('N0677')
        self.assertIsNone(station.forcast_to().tzinfo)
        self.assertIsNone(station.forcast_from().tzinfo)
        station.forecast()

    def test_forecast(self):
        station = WeatherStation('N0677')
        tomorrow = datetime.now()+ timedelta(days=1)
        forcast = station.forecast(tomorrow)
        self.assertEqual(tomorrow.strftime("%d.%m.%Y %H:%M"), forcast.time.strftime("%d.%m.%Y %H:%M"))
        self.assertIsNone(forcast.time.tzinfo)
        self.assertIsNotNone(forcast.time_utc.tzinfo)
        #print("utc:    " + str(forcast.time_utc))
        #print("local:  " + str(forcast.time))




if __name__ == '__main__':
    unittest.main()


import unittest
import gzip
import tempfile
from os import path
from pvpower.traindata import TrainSampleLog, LabelledWeatherForecast
from datetime import datetime



class TestTrainSampleLog(unittest.TestCase):

    def test_uct_format(self):
        with tempfile.TemporaryDirectory() as tmpdirname:
            log = TrainSampleLog(tmpdirname)
            dt1 = datetime.strptime("2021.11.30T13:30", "%Y.%m.%dT%H:%M")
            log.append(LabelledWeatherForecast(dt1, 100, 200, 300, 400, 500, 99))

            with gzip.open(path.join(tmpdirname, TrainSampleLog.FILENAME), 'rb') as f:
                lines = [raw_line.decode('UTF-8').strip() for raw_line in f.readlines()]
                self.assertEqual(1, len(lines))
                self.assertEqual('2021-11-30T12:30:00+00:00;99;100;200;300;400;500', lines[0])

            traindata = log.all()
            self.assertEqual(1, len(traindata.samples))
            self.assertEqual(dt1, traindata.samples[0].time)


if __name__ == '__main__':
    unittest.main()


import logging
from dataclasses import dataclass
from datetime import datetime
from abc import ABC, abstractmethod
from sklearn import svm
from typing import List
from pvpower.weather_forecast import WeatherForecast
from pvpower.traindata import LabelledWeatherForecast, TrainData



class Vectorizer(ABC):

    def _scale(self, value: int, max_value: int, digits=1) -> float:
        if value == 0:
            return 0
        else:
            return round(value * 100 / max_value, digits)

    @abstractmethod
    def vectorize(self, sample: WeatherForecast) -> List[float]:
        pass



class CoreVectorizer(Vectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        vectorized = [self._scale(sample.time_utc.month, 12),
                      self._scale(sample.time_utc.hour, 24),
                      self._scale(sample.irradiance, 1000)]
        return vectorized

    def __str__(self):
        return "CoreVectorizer"


class SunshineVectorizer(Vectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        vectorized = [self._scale(sample.time_utc.month, 12),
                      self._scale(sample.time_utc.hour, 24),
                      self._scale(sample.sunshine, 5000)]
        return vectorized

    def __str__(self):
        return "SunshineVectorizer"


class SushinePlusCloudCoverVectorizer(SunshineVectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        return super().vectorize(sample) +  [self._scale(sample.cloud_cover, 200)]

    def __str__(self):
        return "Sunshine+CloudVectorizer"



class PlusVisibilityVectorizer(CoreVectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        return super().vectorize(sample) +  [self._scale(sample.visibility, 50000)]

    def __str__(self):
        return "Core+VisibilityVectorizer"


class PlusSunshineVectorizer(CoreVectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        return super().vectorize(sample) +  [self._scale(sample.sunshine, 5000)]

    def __str__(self):
        return "Core+SunshineVectorizer"


class PlusCloudCoverVectorizer(CoreVectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        return super().vectorize(sample) +  [self._scale(sample.cloud_cover, 200)]

    def __str__(self):
        return "Core+CloudVectorizer"


class PlusVisibilitySunshineVectorizer(CoreVectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        return super().vectorize(sample) +  [self._scale(sample.visibility, 50000),
                                             self._scale(sample.sunshine, 5000)]

    def __str__(self):
        return "Core+Visibility+SunshineVectorizer"


class PlusVisibilityCloudCoverVectorizer(CoreVectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        return super().vectorize(sample) +  [self._scale(sample.visibility, 50000),
                                             self._scale(sample.cloud_cover, 200)]

    def __str__(self):
        return "Core+Visibility+CloudVectorizer"


class PlusVisibilityFogCloudCoverVectorizer(CoreVectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        return super().vectorize(sample) +  [self._scale(sample.visibility, 50000),
                                             self._scale(sample.cloud_cover, 200),
                                             self._scale(sample.probability_for_fog, 100)]

    def __str__(self):
        return "Core+Visibility+Fog+CloudVectorizer"


class FullVectorizer(CoreVectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        return super().vectorize(sample) +  [self._scale(sample.visibility, 50000),
                                             self._scale(sample.cloud_cover, 200),
                                             self._scale(sample.probability_for_fog, 100),
                                             self._scale(sample.sunshine, 5000)]

    def __str__(self):
        return "FullVectorizer"


@dataclass(frozen=True)
class TrainReport:
    samples: List[LabelledWeatherForecast]


class Estimator(ABC):

    @property
    @abstractmethod
    def variant(self) -> str:
        pass

    @abstractmethod
    def date_last_train(self) -> datetime:
        pass

    @abstractmethod
    def retrain(self, train_data: TrainData) -> TrainReport:
        pass

    @abstractmethod
    def predict(self, sample: WeatherForecast) -> int:
        pass


class SVMEstimator(Estimator):

    def __init__(self, vectorizer: Vectorizer):
        self.__clf = svm.SVC(kernel='poly') # it seems that the SVM approach produces good predictions. refer https://www.sciencedirect.com/science/article/pii/S136403212200274X?via%3Dihub and https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.221.4021&rep=rep1&type=pdf
        self.__vectorizer = vectorizer
        self.__date_last_train = datetime.fromtimestamp(0)
        self.__num_samples_last_train = 0
        self.__num_covered_days_last_train = 0

    @property
    def variant(self) -> str:
        return type(self.__vectorizer).__name__

    def date_last_train(self) -> datetime:
        return self.__date_last_train

    def retrain(self, train_data: TrainData) -> TrainReport:
        samples = train_data.samples
        feature_vector_list = [self.__vectorizer.vectorize(sample) for sample in samples]
        label_list = [sample.power_watt for sample in samples]
        if len(set(label_list)) > 1:
            self.__clf.fit(feature_vector_list, label_list)
            self.__date_last_train = datetime.now()
            self.__num_samples_last_train = len(samples)
            self.__num_covered_days_last_train = len(set([sample.time.strftime("%Y.%m.%d") for sample in samples]))
            logging.debug("estimator has been trained " + str(self))
        else:
            logging.debug("estimator can not be trained. Insufficient train data")
        return TrainReport(samples)

    def predict(self, sample: WeatherForecast) -> int:
        if sample.irradiance > 0:   # shortcut. no irradiance means no power
            if self.__num_samples_last_train < 1:
                logging.warning("estimator has not been trained (insufficient train data available). returning 0")
                return 0
            else:
                feature_vector = self.__vectorizer.vectorize(sample)
                predicted = int(self.__clf.predict([feature_vector])[0])
                logging.debug(str(predicted) + " watt predicted for " + str(sample) + " (features: " + str(feature_vector) + ")")
                if predicted >= 0:
                    return predicted
                else:
                    logging.debug("predicted value is " + str(predicted) + " correct them to 0")
                    return 0
        else:
            return 0

    def __str__(self):
        return "SVMEstimator(vectorizer=" + str(self.__vectorizer) + "; trained with " + str(self.__num_samples_last_train) + " samples; age " + str(datetime.now() - self.date_last_train()) + "; time range: " + str(self.__num_covered_days_last_train) + " days)"




import logging
import pickle
from os import path, makedirs
from os.path import exists
from appdirs import user_cache_dir, site_data_dir
from threading import Thread
from datetime import datetime, timedelta
from typing import Optional
from pvpower.weather_forecast import WeatherStation, WeatherForecast
from pvpower.traindata import LabelledWeatherForecast, TrainSampleLog, TrainData
from pvpower.estimator import Estimator, SVMEstimator, TrainReport, FullVectorizer


class ValueRecorder:

    def __init__(self):
        self.start_time = datetime.strptime(datetime.now().strftime("%d.%m.%Y %H") + ":00", "%d.%m.%Y %H:%M")
        self.end_time = self.start_time + timedelta(minutes=60)
        self.__power_values = []
        #logging.debug("value recorder created (" + str(self) + ")")

    def empty(self) -> bool:
        return len(self.__power_values) == 0

    def is_expired(self):
        return datetime.now() >= self.end_time

    def add(self, value: int):
        self.__power_values.append(value)
        #logging.debug("record added to value recorder (" + str(self) + ")")

    @property
    def average(self) -> Optional[int]:
        if len(self.__power_values) == 0:
            return None
        else:
            return int(round(sum(self.__power_values) / len(self.__power_values), 0))

    def __str__(self):
        return self.start_time.strftime("%Y.%m.%d %H:%M") + " -> " + self.end_time.strftime("%Y.%m.%d %H:%M") + "  average power: " + str(self.average) + " num probes: " + str(len(self.__power_values))



class PersistentEstimator(Estimator):

    def __init__(self, default_estimator: Estimator):
        cache_dir = user_cache_dir("pv_forecast", appauthor=False)
        if not path.exists(cache_dir):
            makedirs(cache_dir)
        self.__filename = path.join(cache_dir, 'Estimator_' + default_estimator.variant + '.p')
        self.__estimator = default_estimator
        if exists(self.__filename):
            try:
                with open(self.__filename, 'rb') as file:
                    self.__estimator = pickle.load(file)
                    logging.debug("estimator " + str(self.__estimator) + " loaded from pickle file (" + self.__filename + ")")
            except Exception as e:
                logging.warning("error occurred loading estimator " + str(e))

    @property
    def variant(self) -> str:
        return self.__estimator.variant

    def date_last_train(self) -> datetime:
        return self.__estimator.date_last_train()

    def predict(self, sample: WeatherForecast) -> int:
        return self.__estimator.predict(sample)

    def retrain(self, train_data: TrainData) -> TrainReport:
        train_report = self.__estimator.retrain(train_data)
        self.__store()
        return train_report

    def __store(self):
        try:
            with open(self.__filename, 'wb') as file:
                pickle.dump(self.__estimator, file)
        except Exception as e:
            pass

    def __str__(self):
        return str(self.__estimator)


class PvPowerForecast:

    def __init__(self, station_id: str, pv_forecast_dir: None, estimator: Estimator = None):
        pv_forecast_dir = pv_forecast_dir if pv_forecast_dir is not None else site_data_dir("pv_forecast", appauthor=False)
        self.train_log = TrainSampleLog(pv_forecast_dir)
        self.__train_value_recorder = ValueRecorder()
        self.weather_forecast_service = WeatherStation(station_id)
        self.__estimator = estimator if estimator is not None else PersistentEstimator(SVMEstimator(FullVectorizer()))
        self.__date_last_retrain_initiated = self.__estimator.date_last_train()
        self.__train_if_old(False)

    def add_current_power_reading(self, real_power: int):
        if self.__train_value_recorder.is_expired():
            try:
                if not self.__train_value_recorder.empty():
                    weather_sample = self.weather_forecast_service.forecast(self.__train_value_recorder.start_time)
                    if weather_sample is not None:
                        annotated_sample = LabelledWeatherForecast.create(weather_sample,
                                                                          self.__train_value_recorder.average,
                                                                          time=self.__train_value_recorder.start_time)
                        logging.info("add train sample " + str(annotated_sample))
                        self.train_log.append(annotated_sample)
            finally:
                self.__train_value_recorder = ValueRecorder()
        self.__train_value_recorder.add(real_power)

    def predict(self, time: datetime) -> Optional[int]:
        sample = self.weather_forecast_service.forecast(time)
        if sample is None:
            logging.info("could not predict power. Reason: no weather forecast data available (requested date time: " + time.strftime("%Y.%m.%d %H:%M") +
                         ". Available date time range: " + self.weather_forecast_service.forcast_from().strftime("%Y.%m.%d %H:%M") +
                         " -> " + self.weather_forecast_service.forcast_to().strftime("%Y.%m.%d %H:%M")  + "). Returning None")
            return None
        else:
            return self.predict_by_weather_forecast(sample)

    def predict_by_weather_forecast(self, sample: WeatherForecast) -> int:
        try:
            return self.__estimator.predict(sample)
        finally:
            self.__train_if_old()

    def __train_if_old(self, background: bool = True):
        if datetime.now() > (self.__date_last_retrain_initiated + timedelta(minutes=3 * 60)):
            self.__date_last_retrain_initiated = datetime.now()
            if background:
                Thread(target=self.__estimator.retrain, daemon=True, args=(self.train_log.all(),)).start()
            else:
                self.__estimator.retrain(self.train_log.all())

    def __str__(self):
        return str(self.__estimator)
